#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Docstring
"""Main Initializatino File
So as to not repeat the same initialization for all the simulations
"""

##############################################################################
# Importing Modules

import numpy as np
import lmfit as lf
import pickle
import copy as cp
import pandas as pd
from cycler import cycler
# import urllib.request

# My Modules
import Functions.Convert as cn

from Functions.GalaxyClass import bar, SPARC
# Functions.Baryons.ExpDisk2D as bar
from Functions.Store import Store  # , LMinterface

from Functions.Logic import Bool

# Matplotlib
import matplotlib as mpl
from matplotlib import mlab
from matplotlib import pyplot as plt
from matplotlib import image as mpimg
from matplotlib.gridspec import GridSpec

# Pandas options
pd.set_option('display.max_rows', None)
pd.set_option('display.max_columns', None)

##############################################################################
# File Data
__author__ = "Nathaniel Starkman"
__credits__ = ["Stacy McGaugh, Frederico Lelli"]
__license__ = "GPL"
__maintainer__ = "Nathaniel Starkman"
__status__ = "Development"


# User-defined paths
pathbase = "Output/Tables/"
RCdatabasepath = "Databases/SPARC/MassModels_Lelli2016c.mrt"
GSdatabasepath = "Databases/SPARC/Table1wbulge.mrt"
bulgebiggerthandisk = True


##############################################################################
#  Parameters

# make parameter array
params = lf.Parameters()

#                name    val  vary   min   max   expr
params.add_many(# Gravitational Constant
                ("G", 4.3e-6, False, None, None, None),
                # disk radius
                ("Rd", 1, False, 1, 1e2, None),
                # central surface brightness
                ("sig0", 1, False, 0, None, None),
                # Mass to light ratio
                ("M2L", .5, False, 1e-5, 50, None))

# Bulge mass to light ratio
params.add('delta', value=0.2, min=1e-5, max=50, vary=True)
if bulgebiggerthandisk is True:
    params.add("M2Lb", value=0., vary=False, min=1e-5, max=50, expr='delta + M2L')
else:
    params.add("M2Lb", value=0., vary=False, min=1e-5, max=50)

params.pretty_print()
print('\n')


##############################################################################
# Cut Selection Criteria

criteria = Store("cut criteria")
criteria.add("Inclination Cut", 30., aka="IncCut",
             ifthis="inc", logic="<", where="GS")
criteria.add("Quality Flag Cut", 3, aka="badQ",
             ifthis="Q", logic="=", where="GS")
criteria.add("Low cent SB", 1, aka="lowSB0",
             ifthis="SB0", logic="<", where="GS")
criteria.add("Dist Error", .2, aka="e_D", ifthis=("e_D", "D"),
             logic=">", where="GS", operation=lambda *arg: arg[0] / arg[1])

##############################################################################
# Making Galaxy

__builtins__.Gal = SPARC("Galaxies", params, criteria,
                         makeprlm={'pathbase': pathbase,
                                   'RCdatabasepath': RCdatabasepath,
                                   'GSdatabasepath': GSdatabasepath})

##############################################################################
# Doing Maximum Disk Fit

from Functions.MaxDisk import Fitting as fit

print('\nfitting...')

if bulgebiggerthandisk is True:
    seedvaldict = {"M2L": 0.5, "delta": 0.2}
    varyparams = [["delta", "M2L"], "delta", "M2L"]
else:
    seedvaldict = {"M2L": 0.5, "M2Lb": 0.7}
    varyparams = [["M2Lb", "M2L"], "M2Lb", "M2L"]

fit.do_fits("Log Normal Best Fit", "ln_fit",
            seedvaldict=seedvaldict,
            varyparams=varyparams,
            minmethods=[["nelder", "leastsq"], "leastsq", "leastsq"],
            wgt_func=fit.lognorm_weight,
            wgt_factor=1.,
            M2Lb_degenerate=True, bulge_wgt_func=fit.bulge_lognorm_weight,
            do_emcee=False)

print('\t ...done')

##############################################################################
# Writing Output Files

RCMRT = Gal['ln_fit'].RCf.loc[:, ['ID', 'R', 'Vobs', 'e_Vobs',
                                  'Vgas',
                                  'Vdisk', 'e_Vdisk', 'SBdisk', 'e_SBdisk',
                                  'Vbul', 'SBbul',
                                  'Vbar', 'e_Vbar']].copy()

sorter = Gal['ln_fit'].GSf.sort_values('L[3.6]').index

RCMRT.index.names = ['ID', '#']
RCMRT = RCMRT.round(decimals=3)
# RCMRT[RCMRT == 0].round(decimals=1)

# RCMRT.to_csv(path_or_buf='RC_MRT.mrt', sep="\t", escapechar='\t', float_format='%3.3f', mode='a+')
with open(r'Output/RC_MRT.mrt', 'w') as f:
    f.write(
"""Table: Mass Models Supplement
=========================================================================================================================================
Byte-by-byte Description of file: RC_MRT.txt
-----------------------------------------------------------------------------------------------------------------------------------------
   Bytes Format    Units        Label   Explanations
-----------------------------------------------------------------------------------------------------------------------------------------
   1-     A15      ---          ID      Galaxy identifier
   16-    F.3      kpc          R       Galactocentric radius
   25-    F.3      km/s         Vobs    Observed circular velocity
   33-    F.3      km/s       e_Vobs    Uncertainty in Vobs (1)
   43-    F.3      km/s         Vgas    Gas velocity contribution (2)
   52-    F.3      km/s         Vdisk   Disk velocity contribution (3)
   61-    F.3      km/s       e_Vdisk   Uncertainty in Vdisk
   71-    F.3      solLum/pc2   SBdisk  Disk surface brightness
   81-    F.3      solLum/pc2 e_SBdisk  Disk surface brightness
   93-    F.3      km/s         Vbul    Bulge velocity contribution (3)
   103-   F.3      solLum/pc2   SBbul   Bulge surface brightness
   116-   F.3      km/s         Vbar    Total baryonic velocity (3)
   127-   F.3      km/s       e_Vbar    Total baryonic velocity
-----------------------------------------------------------------------------------------------------------------------------------------
Note (1): Random error due to non-circular motions and/or kinematic asymmetries.
It does not include systematic uncertainties due to inclination corrections.
Note (2): Vgas includes a factor 1.33 to account for cosmological helium.
Note (3): Vdisk and Vbul are given for their maximized M/L [solMass/solLum] at [3.6] in Galaxy Sample Supplement table.
-----------------------------------------------------------------------------------------------------------------------------------------\n""")

    f.write('{col[0]:<15}{col[1]:<9}{col[2]:<8}{col[3]:<10}{col[4]:<9}{col[5]:<9}{col[6]:<10}{col[7]:<10}{col[8]:<12}{col[9]:<10}{col[10]:<13}{col[11]:<11}{col[12]:<5}\n'.format(col=RCMRT.columns.values, num='#'))

    for gal in sorter:
        for row in RCMRT.loc[gal, :].itertuples():
            e_Vdisk = row.e_Vdisk

            f.write('{1:<15}{0.R:<9.2f}{0.Vobs:<9}{0.e_Vobs:<9}{0.Vgas:<9}{0.Vdisk:<10}{e_Vdisk:<9}{0.SBdisk:<11}{0.e_SBdisk:<11}{0.Vbul:<11}{0.SBbul:<12}{0.Vbar:<12}{0.e_Vbar:<4}\n'.format(row, gal, e_Vdisk=e_Vdisk))


GSMRT = Gal['ln_fit'].GSf.loc[:, ['M2L', 'e_M2L', 'M2Lb', 'e_M2Lb', 'T', 'D','e_D', 'inc', 'e_inc', 'L[3.6]', 'e_L[3.6]',
                                  'Lbulge',
                                  'Reff', 'Rp', 'SBeff', 'e_SBeff', 'Rdisk', 'SB0', 'e_SB0', 'Mbar', 'e_Mbar', 'SBbar',
                                  'e_SBbar', 'Vp', 'e_Vp', 'Vb', 'e_Vb']].copy()
GSMRT.sort_values('L[3.6]', inplace=True)
GSMRT = GSMRT.round(decimals=3)

with open(r'Output/GS_MRT.mrt', 'w') as f:
    f.write(
"""Table: Galaxy Sample Supplement
================================================================================
Byte-by-byte Description of file: GS_MRT.mrt
--------------------------------------------------------------------------------
   Bytes Format    Units           Label   Explanations
--------------------------------------------------------------------------------
  1-     A15        ---            ID      Galaxy Name
  16-    F.2     SolMass/solLum    M2L     Mass-to-Light Ratio
  24-    F.2     SolMass/solLum  e_M2L     Unvertainty in M2L
  34-    F.2     SolMass/solLum    M2Lb     Bulge Mass-to-Light Ratio
  43-    A6         ---            T       Hubble Type (1)
  49-    F.2     Mpc               D       Distance
  57-    F.2     Mpc             e_D       Mean error on D
  67-    F.1     deg               Inc     Inclination
  74-    F.1     deg             e_Inc     Mean error on Inc
  83-    F.3     10+9solLum        L[3.6]  Total Luminosity at [3.6]
  92-    F.3     10+9solLum      e_L[3.6]  Mean error on L[3.6]
  103-   F.3     10+9solLum        Lbulge  Bulge Luminosity at [3.6]
  113-   F.2     kpc               Reff    Effective Radius at [3.6]
  122-   F.2     kpc               Rp      Radius of peak Vbar (5)
  131-   F.3     solLum/pc2        SBeff   Effective Surface Brightness at [3.6]
  141-   F.3     solLum/pc2      e_SBeff   Uncertainty in SBeff
  153-   F.2     kpc               Rdisk   Disk Scale Length at [3.6]
  163-   F.3     solLum/pc2        SB0     Disk Central Surface Brightness at [3.6]
  173-   F.3     solLum/pc2      e_SB0     Uncertainty in SBdisk
  186-   F.2     10+9solMass/pc2   Mbar    Baryonic mass (6)
  198-   F.3     10+9solMass/pc2 e_Mbar    Uncertainty in Mbar (6)
  210-   F.3     solLum/pc2        SBbar   Baryonic Surface Brightness at [3.6] (6)
  222-   F.3     solLum/pc2      e_SBbar   Uncertainty in SBbar (6)
  236-   F.2     km/s              Vp      Vobs at Rp (6)
  243-   F.2     km/s            e_Vp      Uncertainty in Vp (6)
  253-   F.3     km/s              Vb      Vbar at Rp (6)
  262-   F.3     km/s            e_Vb      Uncertainty in Vbar (6)
--------------------------------------------------------------------------------
Note (1):
 0 = S0,  1 = Sa, 2 = Sab, 3 = Sb, 4 = Sbc, 5 = Sc,
 6 = Scd, 7 = Sd, 8 = Sdm, 9 = Sm, 10 = Im, 11 = BCD
Note (2):
 1 = Hubble-Flow assuming H0=73 km/s/Mpc and correcting for Virgo-centric infall
 2 = Magnitude of the Tip of the Red Giant Branch
 3 = Cepheids Magnitude-Period Relation
 4 = Ursa Major Cluster of Galaxies
 5 = Supernovae Light Curve
Note (5):
 Values given in Mass Models Supplement table
--------------------------------------------------------------------------------\n""")

    f.write(
      ('{name:<15}{col[0]:<8}{col[1]:<10}{col[2]:<9}{col[4]:<6}{col[5]:<8}'
        '{col[6]:<10}{col[7]:<7}{col[8]:<9}{col[9]:<9}{col[10]:<11}'
        '{col[11]:<10}{col[12]:<9}{col[13]:<9}{col[14]:<10}{col[15]:<12}'
        '{col[16]:<10}{col[17]:<10}{col[18]:<13}{col[19]:<12}{col[20]:<12}'
        '{col[21]:<12}{col[22]:<14}{col[23]:<7}{col[24]:<10}{col[25]:<9}'
        '{col[26]:<8}\n'.format(col=GSMRT.columns.values, name='ID')
      ))
    for row in GSMRT.itertuples():
        f.write(
          ('{0.Index:<15}{0.M2L:<9.2f}{0.e_M2L:<9}{0.M2Lb:<9}{0.T:<6}{0.D:<9}'
           '{0.e_D:<9}{0.inc:<8}{0.e_inc:<8}{0._10:<11}{0._11:<9}'
           '{0.Lbulge:<10}{0.Reff:<9}{0.Rp:<9}{0.SBeff:<12}{0.e_SBeff:<11}'
           '{0.Rdisk:<9}{0.SB0:<12}{0.e_SB0:<11}{1:<12.2f}{2:<12.3f}'
           '{0.SBbar:<14}{0.e_SBbar:<12}{0.Vp:<9}{0.e_Vp:<8}{0.Vb:<10}'
           '{0.e_Vb:<10}\n'.format(row, row.Mbar / 1e9, row.e_Mbar / 1e9)
          ))


print('MRT data tables completed\n\n')

#######################################################################################

method = 'ln_fit'
methodlabel = 'fit'

for ID in Gal.IDs:

    # if any(Gal.fid.RCf.loc[ID, "Vbul"] != 0):
    #     print(ID, ("{:.2}".format(Gal[method].GSf.loc[ID, "M2L"]),
    #                "{:.2}".format(Gal[method].GSf.loc[ID, "M2Lb"])), end="\n")
    # else:
    #     print(ID, end="\n")

    galx = Gal.galx[ID]

    # Figure & Axes Properties
    fig = plt.figure(figsize=(5, 5))
    gs = GridSpec(1, 1, top=.9)
    ax = plt.subplot(gs[0])

    ax.errorbar(Gal.fid.RCf.loc[ID, "R"], Gal.fid.RCf.loc[ID, "Vobs"],
                yerr=Gal.fid.RCf.loc[ID, "e_Vobs"],
                fmt='b.', ecolor='b')

    # Fit Plot
    met, *_ = ax.errorbar(Gal[method].RCf.loc[ID, "R"],
                          Gal[method].RCf.loc[ID, "Vbar"],
                          yerr=Gal[method].RCf.loc[ID, "e_Vbar"],
                          label=("{:.2}".format(Gal[method].GSf.loc[ID, "M2L"]),
                                 "{:.2}".format(Gal[method].GSf.loc[ID, "M2Lb"])))

    # Fit Plot Components
    ax.errorbar(Gal[method].RCf.loc[ID, "R"],
                Gal[method].RCf.loc[ID, "Vdisk"],
                yerr=Gal[method].RCf.loc[ID, "e_Vdisk"],
                label="Vdisk")
    if any(Gal[method].RCf.loc[ID, "Vbul"] > 0):
        ax.errorbar(Gal[method].RCf.loc[ID, "R"],
                    Gal[method].RCf.loc[ID, "Vbul"],
                    yerr=Gal[method].RCf.loc[ID, "e_Vbul"],
                    label="Vbul")
    ax.errorbar(Gal[method].RCf.loc[ID, "R"],
                Gal[method].RCf.loc[ID, "Vgas"],
                yerr=0,
                label="Vgas")

    if Gal[method].GSf.loc[ID, "Lbulge"] > 0:
        weight = fit.bulge_lognorm_weight
    else:
        weight = fit.lognorm_weight

    # Weighting
    ax.plot(Gal[method].RCf.loc[ID, "R"],
               (weight(None, ID) * Gal[method].GSf.loc[ID, "Vp"] /
                max(fit.lognorm_weight(None, ID))),
               "k--", alpha=.2, label=r"lognorm wgt (scaled)")

    # Rp
    ax.axvline(Gal[method].GSf.loc[ID, "Rp"],
               c="k", ls="-.", alpha=.3, label="Rp")
    ax.axvline(2.2 * Gal[method].GSf.loc[ID, "Rdisk"],
               c="k", ls=":", alpha=.3, label="2.2Rd")

    ax.set_ylabel(r"V [$\frac{\mathrm{km}}{\mathrm{s}}$]")

    ax.set_xlabel("R [kpc]")
    ax.locator_params(axis="y", nbins=4)

    handles, labels = np.array(ax.get_legend_handles_labels())

    if any(Gal[method].RCf.loc[ID, "Vbul"] > 0):
        ind = [0, 1, 2, 3, 4, 5, 6, 7]  # FIXME
    else:
        ind = [0, 4, 5, 6, 2, 3, 1]  # rearranges handles and labels

    fig.legend(handles[ind], labels[ind], loc='upper center', ncol=4)

    # Saving
    try:
        fig.savefig("Output/images/MaxDisk/test_plots/{}.pdf".format(ID),
                    format="pdf", dpi=1000)
    except:
        print("\n", ID, " failed", "\n")

    plt.close(fig);

print("\n\nplots done")


##############################################################################
# End File Cleanup
del params, pathbase, RCdatabasepath, GSdatabasepath, RCMRT, sorter, GSMRT
